
//
// Copyright (C) 2022 David Carpenter
// 
// Permission is hereby granted, free of charge, 
// to any person obtaining a copy of this software 
// and associated documentation files (the "Software"), 
// to deal in the Software without restriction, 
// including without limitation the rights to use, 
// copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, 
// subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be 
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

// version 1.0.1.1
// *canonicalize path argument. (convert . to an absolute path)
// *remove trailing \ in path argument
// *disable recursion option
// *convert forward slashes (/) to backslashes (\).
// *hash algorithm switches. (if none specified use all, eg: /sha1 /md5 /sfv)
// *multiple threads (one thread per algorthim)
// *stat output
// *LFN support
// *sha512 support
// version 1.0.0.0
// initial release.

// TODO:
// option to save file size in .md5, .sha1 and .sha256 output.
// to add an option to regenerate sidecar files.
// add missing files to the sidecar file.
// track timestamps and update changed files.


// compiler options
#pragma warning(disable : 4311) // type cast void * to unsigned int
#pragma warning(disable : 4312) // type cast unsigned int to void *
#pragma warning(disable : 4244) // warning C4244: 'argument' : conversion from 'LONG_PTR' to 'LONG', possible loss of data
#pragma warning(disable : 4201) // nonstandard extension used : nameless struct/union
#pragma warning(disable : 4313) // 'debug_printf' : '%x' in format string conflicts with argument 2 of type 'line_t *'
#pragma warning(disable : 4996) // deprecation
#pragma warning(disable : 4701) // potentially uninitialized local variable 'last_stretch_mode' used

#include <stdio.h>
#include <windows.h>
#include <Wincrypt.h>
#include <shlwapi.h>

#define os_byte_swap_dword(dword) _byteswap_ulong(dword)
#define os_byte_swap_word(word) _byteswap_ushort(word)
#define os_byte_swap_qword(qword) _byteswap_uint64(qword)

#define _FILE_CONTENT_CRC32_KCRCPOLY	0xEDB88320
#define _FILE_CONTENT_CRC32_INIT		0xffffffff
#define _FILE_CONTENT_CRC32_UPDATE_BYTE_2(crc, b) (voidhash_crc32_table[((crc) ^ (b)) & 0xFF] ^ ((crc) >> 8))
#define VOIDHASH_BUF_SIZE				(8*1024*1024)

#define VERSION_MAJOR						1
#define VERSION_MINOR						0
#define VERSION_REVISION					1
#define VERSION_BUILD						2
#define VERSION_YEAR						2022

#define VOIDHASH_ERROR_SUCCESS				0
#define VOIDHASH_ERROR_OUT_OF_MEMORY		1
#define VOIDHASH_ERROR_BAD_PARAMETERS		2
#define VOIDHASH_ERROR_CRYPT				3
#define VOIDHASH_ERROR_OPEN_FILE			4
#define VOIDHASH_ERROR_WRITE_FILE			5

#define VOIDHASH_STRING_MAXLEN			(32768)

static wchar_t *string_copy(wchar_t *d,const wchar_t *p);
static void *mem_alloc(uintptr_t size);
static void mem_free(void *p);
static void voidhash_usage(void);
static HANDLE voidhash_create_thread(LPTHREAD_START_ROUTINE lpStartAddress,void *param);
static uintptr_t string_len(const wchar_t *s);
static wchar_t *string_cat(wchar_t *d,const wchar_t *p);
static void string_copy_lfn(wchar_t *d,const wchar_t *p);
static void voidhash_get_hashes(const wchar_t *filename,DWORD *pcrc,BYTE *pmd5,BYTE *psha1,BYTE *psha256,BYTE *psha512);
static void voidhash_fprintf(HANDLE h,char *format,...);
static int voidhash_file_with_size_exists(const wchar_t *filename);

static DWORD *voidhash_crc32_table = 0;
BYTE *voidhash_buf = 0;
wchar_t *voidhash_lfn = 0;
DWORD voidhash_buf_size = VOIDHASH_BUF_SIZE;
int voidhash_sfv = 0;
int voidhash_md5 = 0;
int voidhash_sha1 = 0;
int voidhash_sha256 = 0;
int voidhash_sha512 = 0;
int voidhash_subfolders = 1;

DWORD crc_value;
HCRYPTHASH md5_hHash;
HCRYPTHASH sha1_hHash;
HCRYPTHASH sha256_hHash;
HCRYPTHASH sha512_hHash;
DWORD voidhash_numread;

HANDLE crc32_wake_event;
HANDLE crc32_complete_event;
HANDLE crc32_thread_handle;
HANDLE md5_wake_event;
HANDLE md5_complete_event;
HANDLE md5_thread_handle;
HANDLE sha1_wake_event;
HANDLE sha1_complete_event;
HANDLE sha1_thread_handle;
HANDLE sha256_wake_event;
HANDLE sha256_complete_event;
HANDLE sha256_thread_handle;
HANDLE sha512_wake_event;
HANDLE sha512_complete_event;
HANDLE sha512_thread_handle;
HANDLE terminate_event;

unsigned __int64 total_size = 0;

// based on Igor Pavlov implementation for 7zip.
static DWORD _file_content_update_crc32(DWORD v,const void *data,uintptr_t size)
{
	const BYTE *p;

	p = (const BYTE *)data;

	for (;size>0 && ((unsigned)(ptrdiff_t)p & 3) != 0; size--, p++)
	{
		v = _FILE_CONTENT_CRC32_UPDATE_BYTE_2(v,*p);
	}

	for (; size >= 4; size -= 4, p += 4)
	{
		v ^= *(const DWORD *)p;
		v = voidhash_crc32_table[0x300 + ((v) & 0xFF)] ^ voidhash_crc32_table[0x200 + ((v >>  8) & 0xFF)] ^ voidhash_crc32_table[0x100 + ((v >> 16) & 0xFF)] ^ voidhash_crc32_table[0x000 + ((v >> 24))];
	}
	
	for (; size > 0; size--, p++)
	{
		v = _FILE_CONTENT_CRC32_UPDATE_BYTE_2(v,*p);
	}
	
	return v;
}

static void _file_content_create_crc32_table(void)
{
	int i;
	
	voidhash_crc32_table = (DWORD *)malloc(256 * 8 * sizeof(DWORD));
		
	for (i=0;i<256;i++)
	{
		DWORD r;
		DWORD j;
		
		r = i;
		
		for (j=0;j<8;j++)
		{
			r = (r >> 1) ^ (_FILE_CONTENT_CRC32_KCRCPOLY & ~((r & 1) - 1));
		}
		
		voidhash_crc32_table[i] = r;
	}
	
	for (;i<256*8;i++)
	{
		DWORD r;
		
		r = voidhash_crc32_table[i - 256];
		voidhash_crc32_table[i] = voidhash_crc32_table[r & 0xFF] ^ (r >> 8);
	}
}

DWORD WINAPI crc32_thread_proc(void *param)
{
	for(;;)
	{
		WaitForSingleObject(crc32_wake_event,INFINITE);
		ResetEvent(crc32_wake_event);
		
//		printf("crc thread awake\n");
		
		if (WaitForSingleObject(terminate_event,0) == WAIT_OBJECT_0)
		{
			break;
		}

		crc_value = _file_content_update_crc32(crc_value,voidhash_buf,voidhash_numread);
		
		SetEvent(crc32_complete_event);
		
//		printf("crc thread sleep\n");
	}

//	printf("crc thread done\n");
	
	return 0;
}

DWORD WINAPI md5_thread_proc(void *param)
{
	for(;;)
	{
		WaitForSingleObject(md5_wake_event,INFINITE);
		ResetEvent(md5_wake_event);
		
//		printf("md5 thread awake\n");
		
		if (WaitForSingleObject(terminate_event,0) == WAIT_OBJECT_0)
		{
			break;
		}

		if (!CryptHashData(md5_hHash,voidhash_buf,voidhash_numread,0))
		{
			printf("CryptHashData\n");
			ExitProcess(VOIDHASH_ERROR_CRYPT);
		}
		
		SetEvent(md5_complete_event);
		
//		printf("md5 thread sleep\n");
	}

//	printf("md5 thread done\n");
	
	return 0;
}

DWORD WINAPI sha1_thread_proc(void *param)
{
	for(;;)
	{
		WaitForSingleObject(sha1_wake_event,INFINITE);
		ResetEvent(sha1_wake_event);
		
//		printf("sha1 thread awake\n");
		
		if (WaitForSingleObject(terminate_event,0) == WAIT_OBJECT_0)
		{
			break;
		}

		if (!CryptHashData(sha1_hHash,voidhash_buf,voidhash_numread,0))
		{
			printf("CryptHashData\n");
			ExitProcess(VOIDHASH_ERROR_CRYPT);
		}
		
		SetEvent(sha1_complete_event);
		
//		printf("sha1 thread sleep\n");
	}

//	printf("sha1 thread done\n");
	
	return 0;
}

DWORD WINAPI sha256_thread_proc(void *param)
{
	for(;;)
	{
		WaitForSingleObject(sha256_wake_event,INFINITE);
		ResetEvent(sha256_wake_event);
		
//		printf("sha256 thread awake\n");
		
		if (WaitForSingleObject(terminate_event,0) == WAIT_OBJECT_0)
		{
			break;
		}

		if (!CryptHashData(sha256_hHash,voidhash_buf,voidhash_numread,0))
		{
			printf("CryptHashData\n");
			ExitProcess(VOIDHASH_ERROR_CRYPT);
		}
		
		SetEvent(sha256_complete_event);
		
//		printf("sha256 thread sleep\n");
	}

//	printf("sha256 thread done\n");
	
	return 0;
}

DWORD WINAPI sha512_thread_proc(void *param)
{
	for(;;)
	{
		WaitForSingleObject(sha512_wake_event,INFINITE);
		ResetEvent(sha512_wake_event);
		
//		printf("sha512 thread awake\n");
		
		if (WaitForSingleObject(terminate_event,0) == WAIT_OBJECT_0)
		{
			break;
		}

		if (!CryptHashData(sha512_hHash,voidhash_buf,voidhash_numread,0))
		{
			printf("CryptHashData\n");
			ExitProcess(VOIDHASH_ERROR_CRYPT);
		}
		
		SetEvent(sha512_complete_event);
		
//		printf("sha512 thread sleep\n");
	}

//	printf("sha512 thread done\n");
	
	return 0;
}

static void voidhash_get_hashes(const wchar_t *filename,DWORD *pcrc,BYTE *pmd5,BYTE *psha1,BYTE *psha256,BYTE *psha512)
{
	HANDLE h;

	string_copy_lfn(voidhash_lfn,filename);
	
	h = CreateFileW(voidhash_lfn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
	
	if (h != INVALID_HANDLE_VALUE)	
	{
		HCRYPTPROV md5_hProv;
		DWORD md5_alg_id;
		DWORD md5_hash_size;
		DWORD md5_dwProvType;
		HCRYPTPROV sha1_hProv;
		DWORD sha1_alg_id;
		DWORD sha1_hash_size;
		DWORD sha1_dwProvType;
		HCRYPTPROV sha256_hProv;
		DWORD sha256_alg_id;
		DWORD sha256_hash_size;
		DWORD sha256_dwProvType;
		HCRYPTPROV sha512_hProv;
		DWORD sha512_alg_id;
		DWORD sha512_hash_size;
		DWORD sha512_dwProvType;
		DWORD cbHash;

		// ALG_SID_MD5
		md5_alg_id = (4 << 13) | (0) | 3;
		md5_hash_size = 16;
		md5_dwProvType = 1;

		// CALG_SHA1
		sha1_alg_id = (4 << 13) | (0) | 4;
		sha1_hash_size = 20;
		sha1_dwProvType = 1;

		// CALG_SHA_256
		sha256_alg_id = (4 << 13) | (0) | 12;
		sha256_hash_size = 32;
		sha256_dwProvType = 24;

		// CALG_SHA_512
		sha512_alg_id = (4 << 13) | (0) | 14;
		sha512_hash_size = 64;
		sha512_dwProvType = 24;
			
	// #define PROV_RSA_FULL	1
	// #define PROV_RSA_AES     24
	// #define CRYPT_VERIFYCONTEXT	0xF0000000
		if (pmd5)
		{
			if (!CryptAcquireContextW(&md5_hProv,NULL,NULL,md5_dwProvType,0xF0000000))
			{
				printf("CryptAcquireContextW\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
			
			if (!CryptCreateHash(md5_hProv,md5_alg_id,0,0,&md5_hHash))
			{
				printf("CryptCreateHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (psha1)		
		{
			if (!CryptAcquireContextW(&sha1_hProv,NULL,NULL,sha1_dwProvType,0xF0000000))
			{
				printf("CryptAcquireContextW\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
			
			if (!CryptCreateHash(sha1_hProv,sha1_alg_id,0,0,&sha1_hHash))
			{
				printf("CryptCreateHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (psha256)
		{
			if (!CryptAcquireContextW(&sha256_hProv,NULL,NULL,sha256_dwProvType,0xF0000000))
			{
				printf("CryptAcquireContextW\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
			
			if (!CryptCreateHash(sha256_hProv,sha256_alg_id,0,0,&sha256_hHash))
			{
				printf("CryptCreateHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (psha512)
		{
			if (!CryptAcquireContextW(&sha512_hProv,NULL,NULL,sha512_dwProvType,0xF0000000))
			{
				printf("CryptAcquireContextW\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
			
			if (!CryptCreateHash(sha512_hProv,sha512_alg_id,0,0,&sha512_hHash))
			{
				printf("CryptCreateHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (pcrc)
		{
			crc_value = _FILE_CONTENT_CRC32_INIT;
		}

		for(;;)
		{
			DWORD numread;
			
			if (!ReadFile(h,voidhash_buf,voidhash_buf_size,&numread,NULL))
			{
				break;
			}
			
			if (!numread)
			{
				// EOF
				break;
			}
			
			voidhash_numread = numread;
			total_size += numread;
			
			if (pmd5)
			{
				ResetEvent(md5_complete_event);
				SetEvent(md5_wake_event);
			}
			
			if (psha1)
			{
				ResetEvent(sha1_complete_event);
				SetEvent(sha1_wake_event);
			}
			
			if (psha256)
			{
				ResetEvent(sha256_complete_event);
				SetEvent(sha256_wake_event);
			}
			
			if (psha512)
			{
				ResetEvent(sha512_complete_event);
				SetEvent(sha512_wake_event);
			}
			
			if (pcrc)
			{
				ResetEvent(crc32_complete_event);
				SetEvent(crc32_wake_event);
			}
			
			// wait..
			
			if (pmd5)
			{
				WaitForSingleObject(md5_complete_event,INFINITE);
			}
			
			if (psha1)
			{
				WaitForSingleObject(sha1_complete_event,INFINITE);
			}
			
			if (psha256)
			{
				WaitForSingleObject(sha256_complete_event,INFINITE);
			}
			
			if (psha512)
			{
				WaitForSingleObject(sha512_complete_event,INFINITE);
			}
			
			if (pcrc)
			{
				WaitForSingleObject(crc32_complete_event,INFINITE);
			}
		}
		
		if (pmd5)
		{
			cbHash = md5_hash_size;

			// #define HP_HASHVAL              0x0002  // Hash value
			if (!CryptGetHashParam(md5_hHash,0x0002,pmd5,&cbHash,0))
			{
				printf("CryptGetHashParam\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (cbHash != md5_hash_size)
			{
				printf("md5_hash_size\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
						
			if (!CryptDestroyHash(md5_hHash))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (!CryptReleaseContext(md5_hProv,0))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (psha1)
		{
			cbHash = sha1_hash_size;

			// #define HP_HASHVAL              0x0002  // Hash value
			if (!CryptGetHashParam(sha1_hHash,0x0002,psha1,&cbHash,0))
			{
				printf("CryptGetHashParam\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (cbHash != sha1_hash_size)
			{
				printf("sha1_hash_size\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
						
			if (!CryptDestroyHash(sha1_hHash))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (!CryptReleaseContext(sha1_hProv,0))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (psha256)
		{
			cbHash = sha256_hash_size;

			// #define HP_HASHVAL              0x0002  // Hash value
			if (!CryptGetHashParam(sha256_hHash,0x0002,psha256,&cbHash,0))
			{
				printf("CryptGetHashParam\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (cbHash != sha256_hash_size)
			{
				printf("sha256_hash_size\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
						
			if (!CryptDestroyHash(sha256_hHash))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (!CryptReleaseContext(sha256_hProv,0))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (psha512)
		{
			cbHash = sha512_hash_size;

			// #define HP_HASHVAL              0x0002  // Hash value
			if (!CryptGetHashParam(sha512_hHash,0x0002,psha512,&cbHash,0))
			{
				printf("CryptGetHashParam\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (cbHash != sha512_hash_size)
			{
				printf("sha512_hash_size\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
						
			if (!CryptDestroyHash(sha512_hHash))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}

			if (!CryptReleaseContext(sha512_hProv,0))
			{
				printf("CryptDestroyHash\n");
				ExitProcess(VOIDHASH_ERROR_CRYPT);
			}
		}
		
		if (pcrc)
		{
			*pcrc = crc_value ^ _FILE_CONTENT_CRC32_INIT;
		}

		CloseHandle(h);
	}
	else
	{
		printf("Open %S\n",voidhash_lfn);
		ExitProcess(VOIDHASH_ERROR_OPEN_FILE);
	}
}

wchar_t *getext(wchar_t *p)
{
	wchar_t *last;
	
	last = L"";
	
	while(*p)
	{
		if (*p == '.')
		{
			last = p;
		}
		
		p++;
	}
	
	return last;
}

int is_good_ext(wchar_t *ext)
{
	if (wcscmp(ext,L".sfv") == 0)
	{
		return 0;
	}
	
	if (wcscmp(ext,L".sha256") == 0)
	{
		return 0;
	}
	
	if (wcscmp(ext,L".sha512") == 0)
	{
		return 0;
	}
	
	if (wcscmp(ext,L".md5") == 0)
	{
		return 0;
	}
	
	if (wcscmp(ext,L".sha1") == 0)
	{
		return 0;
	}
	
	return 1;
}

static void voidhash_fprintf(HANDLE h,char *format,...)
{
	va_list argptr;
	DWORD written;
	uintptr_t len;
			
	va_start(argptr,format);
		
	vsprintf((char *)voidhash_lfn,format,argptr);
	
	len = strlen((char *)voidhash_lfn);
	
	if (!WriteFile(h,voidhash_lfn,len,&written,NULL))
	{
		printf("WriteFile\n");
		ExitProcess(VOIDHASH_ERROR_WRITE_FILE);
	}
	
	if (written != len)
	{
		printf("WriteFile\n");
		ExitProcess(VOIDHASH_ERROR_WRITE_FILE);
	}
	
	va_end(argptr);
}

void parse(wchar_t *path,wchar_t *parent)
{
	wchar_t *buf;
	WIN32_FIND_DATAW fd;
	HANDLE h;
	HANDLE outsfv;
	HANDLE outmd5;
	HANDLE outsha1;
	HANDLE outsha256;
	HANDLE outsha512;
	
	buf = mem_alloc(VOIDHASH_STRING_MAXLEN * sizeof(wchar_t));
	
	outsfv = INVALID_HANDLE_VALUE;
	outmd5 = INVALID_HANDLE_VALUE;
	outsha1 = INVALID_HANDLE_VALUE;
	outsha256 = INVALID_HANDLE_VALUE;
	outsha512 = INVALID_HANDLE_VALUE;
	
	_swprintf(buf,L"%s\\*.*",path);
	
	string_copy_lfn(voidhash_lfn,buf);
						
	h = FindFirstFileW(voidhash_lfn,&fd);
	if (h != INVALID_HANDLE_VALUE)
	{
		for(;;)
		{
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				if ((wcscmp(fd.cFileName,L"..") != 0) && (wcscmp(fd.cFileName,L".") != 0))
				{
					if (voidhash_subfolders)
					{
						HANDLE fh;
						
						_swprintf(buf,L"%s\\%s",path,fd.cFileName);
						parse(buf,fd.cFileName);
						
						string_copy_lfn(voidhash_lfn,buf);

						// restore date modified.
						fh = CreateFile(voidhash_lfn,FILE_WRITE_ATTRIBUTES,0,0,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
						if (fh != INVALID_HANDLE_VALUE)
						{
	//					printf("set ft %s\n",fd.cFileName);
							if (!SetFileTime(fh,NULL,NULL,&fd.ftLastWriteTime))
							{
								printf("FAILED TO SET FILETIME %s\n",fd.cFileName);
							}
							
							CloseHandle(fh);
						}
						else
						{
							printf("invalid file handle %s\n",buf);
						}
					}
				}
			}
			else
			if (parent)
			{
				wchar_t *ext;
			
				ext = getext(fd.cFileName);
				
				if (is_good_ext(ext))
				{
					DWORD crc;
					BYTE md5[16];
					BYTE sha1[20];
					BYTE sha256[32];
					BYTE sha512[64];
					
					if (voidhash_sfv)
					{
						if (outsfv == INVALID_HANDLE_VALUE)
						{
							_swprintf(buf,L"%s\\%s.sfv",path,parent);

							string_copy_lfn(voidhash_lfn,buf);
							
							if (!voidhash_file_with_size_exists(voidhash_lfn))
							{
								outsfv = CreateFileW(voidhash_lfn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
							}
						}
					}

					if (voidhash_md5)
					{
						if (outmd5 == INVALID_HANDLE_VALUE)
						{
							_swprintf(buf,L"%s\\%s.md5",path,parent);
							
							string_copy_lfn(voidhash_lfn,buf);
							
							if (!voidhash_file_with_size_exists(voidhash_lfn))
							{
								outmd5 = CreateFileW(voidhash_lfn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
							}
						}
					}

					if (voidhash_sha1)
					{
						if (outsha1 == INVALID_HANDLE_VALUE)
						{
							_swprintf(buf,L"%s\\%s.sha1",path,parent);
							
							string_copy_lfn(voidhash_lfn,buf);
							
							if (!voidhash_file_with_size_exists(voidhash_lfn))
							{
								outsha1 = CreateFileW(voidhash_lfn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
							}
						}
					}

					if (voidhash_sha256)
					{
						if (outsha256 == INVALID_HANDLE_VALUE)
						{
							_swprintf(buf,L"%s\\%s.sha256",path,parent);
							
							string_copy_lfn(voidhash_lfn,buf);
							
							if (!voidhash_file_with_size_exists(voidhash_lfn))
							{
								outsha256 = CreateFileW(voidhash_lfn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
							}
						}
					}

					if (voidhash_sha512)
					{
						if (outsha512 == INVALID_HANDLE_VALUE)
						{
							_swprintf(buf,L"%s\\%s.sha512",path,parent);
							
							string_copy_lfn(voidhash_lfn,buf);
							
							if (!voidhash_file_with_size_exists(voidhash_lfn))
							{
								outsha512 = CreateFileW(voidhash_lfn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
							}
						}
					}

					if ((outsfv != INVALID_HANDLE_VALUE) || (outmd5 != INVALID_HANDLE_VALUE) || (outsha1 != INVALID_HANDLE_VALUE) || (outsha256 != INVALID_HANDLE_VALUE) || (outsha512 != INVALID_HANDLE_VALUE))
					{
						_swprintf(buf,L"%s\\%s",path,fd.cFileName);

						printf("%S\n",fd.cFileName);
	
						voidhash_get_hashes(buf,(outsfv != INVALID_HANDLE_VALUE) ? &crc : NULL,(outmd5 != INVALID_HANDLE_VALUE) ? md5 : NULL,(outsha1 != INVALID_HANDLE_VALUE) ? sha1 : NULL,(outsha256 != INVALID_HANDLE_VALUE) ? sha256 : NULL,(outsha512 != INVALID_HANDLE_VALUE) ? sha512 : NULL);
					}
					
					if (outsfv != INVALID_HANDLE_VALUE)
					{
						WideCharToMultiByte(CP_UTF8,0,fd.cFileName,-1,(BYTE *)buf,MAX_PATH,0,NULL);
					
						voidhash_fprintf(outsfv,"%s %08x\r\n",(BYTE *)buf,crc);
					}

					if (outmd5 != INVALID_HANDLE_VALUE)
					{
						WideCharToMultiByte(CP_UTF8,0,fd.cFileName,-1,(BYTE *)buf,MAX_PATH,0,NULL);
					
						voidhash_fprintf(outmd5,"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x %s\r\n",md5[0],md5[1],md5[2],md5[3],md5[4],md5[5],md5[6],md5[7],md5[8],md5[9],md5[10],md5[11],md5[12],md5[13],md5[14],md5[15],(BYTE *)buf);
					}

					if (outsha1 != INVALID_HANDLE_VALUE)
					{
						WideCharToMultiByte(CP_UTF8,0,fd.cFileName,-1,(BYTE *)buf,MAX_PATH,0,NULL);
					
						voidhash_fprintf(outsha1,"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x %s\r\n",sha1[0],sha1[1],sha1[2],sha1[3],sha1[4],sha1[5],sha1[6],sha1[7],sha1[8],sha1[9],sha1[10],sha1[11],sha1[12],sha1[13],sha1[14],sha1[15],sha1[16],sha1[17],sha1[18],sha1[19],(BYTE *)buf);
					}

					if (outsha256 != INVALID_HANDLE_VALUE)
					{
						WideCharToMultiByte(CP_UTF8,0,fd.cFileName,-1,(BYTE *)buf,MAX_PATH,0,NULL);
					
						voidhash_fprintf(outsha256,"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x %s\r\n",sha256[0],sha256[1],sha256[2],sha256[3],sha256[4],sha256[5],sha256[6],sha256[7],sha256[8],sha256[9],sha256[10],sha256[11],sha256[12],sha256[13],sha256[14],sha256[15],sha256[16],sha256[17],sha256[18],sha256[19],sha256[20],sha256[21],sha256[22],sha256[23],sha256[24],sha256[25],sha256[26],sha256[27],sha256[28],sha256[29],sha256[30],sha256[31],(BYTE *)buf);
					}

					if (outsha512 != INVALID_HANDLE_VALUE)
					{
						WideCharToMultiByte(CP_UTF8,0,fd.cFileName,-1,(BYTE *)buf,MAX_PATH,0,NULL);
					
						voidhash_fprintf(outsha512,"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x %s\r\n",sha512[0],sha512[1],sha512[2],sha512[3],sha512[4],sha512[5],sha512[6],sha512[7],sha512[8],sha512[9],sha512[10],sha512[11],sha512[12],sha512[13],sha512[14],sha512[15],sha512[16],sha512[17],sha512[18],sha512[19],sha512[20],sha512[21],sha512[22],sha512[23],sha512[24],sha512[25],sha512[26],sha512[27],sha512[28],sha512[29],sha512[30],sha512[31],sha512[32],sha512[33],sha512[34],sha512[35],sha512[36],sha512[37],sha512[38],sha512[39],sha512[40],sha512[41],sha512[42],sha512[43],sha512[44],sha512[45],sha512[46],sha512[47],sha512[48],sha512[49],sha512[50],sha512[51],sha512[52],sha512[53],sha512[54],sha512[55],sha512[56],sha512[57],sha512[58],sha512[59],sha512[60],sha512[61],sha512[62],sha512[63],(BYTE *)buf);
					}
				}
			}
			
			if (!FindNextFile(h,&fd)) break;
		}
		
		FindClose(h);
	}
	
	if (outsfv != INVALID_HANDLE_VALUE)
	{
		CloseHandle(outsfv);
	}
	
	if (outmd5 != INVALID_HANDLE_VALUE)
	{
		CloseHandle(outmd5);
	}
	
	if (outsha1 != INVALID_HANDLE_VALUE)
	{
		CloseHandle(outsha1);
	}
	
	if (outsha256 != INVALID_HANDLE_VALUE)
	{
		CloseHandle(outsha256);
	}

	if (outsha512 != INVALID_HANDLE_VALUE)
	{
		CloseHandle(outsha512);
	}
	
	mem_free(buf);
}

static void *mem_alloc(uintptr_t size)
{
	void *p;
	
	p = malloc(size);
	if (!p)
	{
		printf("Out of memory %zu\n",size);
		
		ExitProcess(VOIDHASH_ERROR_OUT_OF_MEMORY);
	}
	
	return p;
}

static void mem_free(void *p)
{
	free(p);
}

static void voidhash_usage(void)
{
	printf("Usage: voidhash.exe [options] <path>\n");
	printf("Example: voidhash.exe -sha256 \"D:\\Backup\"\n");
	printf("\n");
	printf("Options:\n");
	printf("   -sfv\n");
	printf("   -md5\n");
	printf("   -sha1\n");
	printf("   -sha256\n");
	printf("   -sha512\n");
	printf("      Specify hash algorithms.\n");
	printf("      Default is: -sfv -md5 -sha1 -sha256\n");
	printf("\n");
	printf("   -bufsize <size>\n");
	printf("      Specify buffer size in bytes\n");
	printf("      Default is: -bufsize 8388608\n");
	printf("\n");
	printf("   -nosubfolders\n");
	printf("      Ignore subfolders\n");
	printf("      Default is to recurse subfolders.\n");
	printf("\n");
	
	ExitProcess(VOIDHASH_ERROR_BAD_PARAMETERS);
}

static wchar_t *string_copy(wchar_t *d,const wchar_t *p)
{	
	uintptr_t run;
	
	run = VOIDHASH_STRING_MAXLEN - 1;
	
	while(run)
	{
		if (!*p)
		{
			break;
		}
		
		*d++ = *p;
		p++;
		run--;
	}
	
	*d = 0;
	
	return d;
}

static wchar_t *string_cat(wchar_t *d,const wchar_t *p)
{	
	uintptr_t run;
	
	run = VOIDHASH_STRING_MAXLEN - 1;
	
	while(run)
	{
		if (!*d)
		{
			break;
		}
		
		d++;
		run--;
	}
	
	while(run)
	{
		if (!*p)
		{
			break;
		}
		
		*d++ = *p;
		p++;
		run--;
	}
	
	*d = 0;
	
	return d;
}

static void string_copy_lfn(wchar_t *d,const wchar_t *p)
{
	uintptr_t len;
	
	len = string_len(p);
	
	if (len < MAX_PATH)
	{
		string_copy(d,p);
	}
	else
	{
		string_copy(d,L"\\\\?\\");
		string_cat(d,p);
	}
}

static uintptr_t string_len(const wchar_t *s)
{
	const wchar_t *p;
	
	p = s;
	
	while(*p)
	{
		p++;
	}
	
	return p - s;
}

static HANDLE voidhash_create_thread(LPTHREAD_START_ROUTINE lpStartAddress,void *param)
{
	DWORD threadid;

	return CreateThread(NULL,0,lpStartAddress,0,0,&threadid);
}

static int voidhash_file_with_size_exists(const wchar_t *filename)
{
	WIN32_FIND_DATA fd;
	
	if (GetFileAttributesEx(filename,GetFileExInfoStandard,&fd))
	{
		if (fd.nFileSizeLow)
		{
			return 1;
		}

		if (fd.nFileSizeHigh)
		{
			return 1;
		}
	}
	
	return 0;
}

void voidhash_main(void)
{
	int argc;
	wchar_t **wargv;
	wchar_t *path;
	int got_hash_type;
	int got_path;
	uintptr_t path_len;
	DWORD tickstart;

	printf("voidhash %d.%d.%d.%d (c) voidtools %d\n",VERSION_MAJOR,VERSION_MINOR,VERSION_REVISION,VERSION_BUILD,VERSION_YEAR);
	
	got_hash_type = 0;
	got_path = 0;
	tickstart = GetTickCount();
	
	path = mem_alloc(VOIDHASH_STRING_MAXLEN*sizeof(wchar_t));
	voidhash_lfn = mem_alloc(VOIDHASH_STRING_MAXLEN*sizeof(wchar_t));
	
	wargv = CommandLineToArgvW(GetCommandLineW(),&argc);

	{
		int i;
		
		for(i=1;i<argc;i++)
		{
			if ((wcsicmp(wargv[i],L"/sfv") == 0) || (wcsicmp(wargv[i],L"-sfv") == 0))
			{
				voidhash_sfv = 1;
				got_hash_type = 1;
			}
			else
			if ((wcsicmp(wargv[i],L"/md5") == 0) || (wcsicmp(wargv[i],L"-md5") == 0))
			{
				voidhash_md5 = 1;
				got_hash_type = 1;
			}
			else
			if ((wcsicmp(wargv[i],L"/sha1") == 0) || (wcsicmp(wargv[i],L"-sha1") == 0))
			{
				voidhash_sha1 = 1;
				got_hash_type = 1;
			}
			else
			if ((wcsicmp(wargv[i],L"/sha256") == 0) || (wcsicmp(wargv[i],L"-sha256") == 0))
			{
				voidhash_sha256 = 1;
				got_hash_type = 1;
			}
			else
			if ((wcsicmp(wargv[i],L"/sha512") == 0) || (wcsicmp(wargv[i],L"-sha512") == 0))
			{
				voidhash_sha512 = 1;
				got_hash_type = 1;
			}
			else
			if ((wcsicmp(wargv[i],L"/bufsize") == 0) || (wcsicmp(wargv[i],L"-bufsize") == 0))
			{
				if (i + 1 < argc)
				{
					i++;
					voidhash_buf_size = _wtoi(wargv[i]);
				}
				else
				{
					voidhash_usage();
				}
			}
			else
			if ((wcsicmp(wargv[i],L"/s") == 0) || (wcsicmp(wargv[i],L"-s") == 0) || (wcsicmp(wargv[i],L"/nosubfolders") == 0) || (wcsicmp(wargv[i],L"-nosubfolders") == 0))
			{
				voidhash_subfolders = 0;
			}
			else
			{
				if (got_path)
				{
					voidhash_usage();
				}
				else
				{
					wchar_t *work_wbuf;
					
					work_wbuf = mem_alloc(VOIDHASH_STRING_MAXLEN * sizeof(wchar_t));
				
					string_copy(work_wbuf,wargv[i]);
					
					// convert '/' to '\\'
					{
						wchar_t *p;
						
						p = work_wbuf;
						
						while(*p)
						{
							if (*p == '/')
							{
								*p = '\\';
							}
							
							p++;
						}
					}
					
					path_len = string_len(work_wbuf);
					
					if (path_len < MAX_PATH)
					{
						PathCanonicalize(path,wargv[i]);
					}
					else
					{
						string_copy(path,wargv[i]);
					}
					
					path_len = string_len(path);
					
					if (path_len)
					{
						if (path[path_len - 1] == '\\')
						{
							path_len--;
							path[path_len] = 0;
						}
					}
					
					got_path = 1;

					mem_free(work_wbuf);
				}
			}
		}
	}
	
	if (!got_path)
	{
		voidhash_usage();
	}
				
	if (!got_hash_type)
	{
		voidhash_sfv = 1;
		voidhash_md5 = 1;
		voidhash_sha1 = 1;
		voidhash_sha256 = 1;
	}

	terminate_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	crc32_wake_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	crc32_complete_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	crc32_thread_handle = voidhash_create_thread(crc32_thread_proc,0);

	md5_wake_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	md5_complete_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	md5_thread_handle = voidhash_create_thread(md5_thread_proc,0);
	
	sha1_wake_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	sha1_complete_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	sha1_thread_handle = voidhash_create_thread(sha1_thread_proc,0);
	
	sha256_wake_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	sha256_complete_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	sha256_thread_handle = voidhash_create_thread(sha256_thread_proc,0);
	
	sha512_wake_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	sha512_complete_event = CreateEvent(NULL,TRUE,FALSE,NULL);
	sha512_thread_handle = voidhash_create_thread(sha512_thread_proc,0);
	
	voidhash_buf = (BYTE *)mem_alloc(voidhash_buf_size);

	_file_content_create_crc32_table();
		
	{
		wchar_t *last;
		wchar_t *p;
		HANDLE h;
		FILETIME dm;
		int got_dm;
		
		last = path;
		p = path;
		got_dm = 0;
		
		while(*p)
		{
			if (*p == '\\')
			{
				last = p + 1;
			}
			
			p++;
		}
		
		string_copy_lfn(voidhash_lfn,path);
		
		h = CreateFile(voidhash_lfn,FILE_WRITE_ATTRIBUTES | FILE_READ_ATTRIBUTES,0,0,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);

		if (h != INVALID_HANDLE_VALUE)
		{
			if (GetFileTime(h,NULL,NULL,&dm))
			{
				got_dm = 1;
			}
		}
		
		parse(path,last);
		
		// restore date modified.
		if (h != INVALID_HANDLE_VALUE)
		{
			if (got_dm)
			{
				if (!SetFileTime(h,NULL,NULL,&dm))
				{
					printf("FAILED TO SET FILETIME\n");
				}
			}
			
			CloseHandle(h);
		}
	}
	
	SetEvent(terminate_event);
	
	SetEvent(sha512_wake_event);
	WaitForSingleObject(sha512_thread_handle,INFINITE);
	CloseHandle(sha512_thread_handle);
	CloseHandle(sha512_complete_event);
	CloseHandle(sha512_wake_event);

	SetEvent(sha256_wake_event);
	WaitForSingleObject(sha256_thread_handle,INFINITE);
	CloseHandle(sha256_thread_handle);
	CloseHandle(sha256_complete_event);
	CloseHandle(sha256_wake_event);

	SetEvent(sha1_wake_event);
	WaitForSingleObject(sha1_thread_handle,INFINITE);
	CloseHandle(sha1_thread_handle);
	CloseHandle(sha1_complete_event);
	CloseHandle(sha1_wake_event);

	SetEvent(md5_wake_event);
	WaitForSingleObject(md5_thread_handle,INFINITE);
	CloseHandle(md5_thread_handle);
	CloseHandle(md5_complete_event);
	CloseHandle(md5_wake_event);

	SetEvent(crc32_wake_event);
	WaitForSingleObject(crc32_thread_handle,INFINITE);
	CloseHandle(crc32_thread_handle);
	CloseHandle(crc32_complete_event);
	CloseHandle(crc32_wake_event);
	
	CloseHandle(terminate_event);

	if (total_size)
	{
		unsigned __int64 total_mb;
		
		total_mb = total_size / (1024 * 1024);
		
		if (total_mb)
		{
			DWORD tickdiff;
			
			tickdiff = (GetTickCount() - tickstart) / 1000;
			if (tickdiff)
			{
				printf("%d MB/s\n",total_mb / tickdiff);
			}
		}
	}

	if (voidhash_buf)
	{
		mem_free(voidhash_buf);
	}

	mem_free(voidhash_lfn);
	mem_free(path);
}

int main(int argc,char **argv)
{
	voidhash_main();
	
	return VOIDHASH_ERROR_SUCCESS;
}
